function varargout = exchange(Sys,Exp,Opt)

% Implements intramolecular chemical exchange using theory from
%   Binsch, J.Am.Chem.Soc. 91, 1304 (1969)
%   Heinzer, Mol.Phys. 22, 167 (1971)
%   Bain, Prog. NMR Spectr. 43, 63 (2003)

% Methods:
% -3 Explicit method, two-site mutual exchange, Matlab accumulation
% -2 Explicit method, two-site mutual exchange, mex accumulation
% -1 Liouville method, two-site mutual exchange, Matlab accumulation
%  0 Liouville method, two-site mutual exchange, mex accumulation
%  1 Liouville method, general n-site exchange, mex accumulation
%  2 Liouville method, general exchange, Matlab accumulation
%  3 Bloch/McConnell, two-site mutual exchange

% Stefan Stoll, 2014

if (nargin<2), Exp = struct; end
if (nargin<3), Opt = struct; end
%if ~isstruct(Sys), error('Spin system structure Sys is missing.'); end
if ~isstruct(Exp), error('Experiment structure Exp is missing.'); end

weight=1;
% Experimental parameters
%------------------------------------------------------------
if ~isfield(Exp,'mwFreq'), Exp.mwFreq = 9.5; end
if ~isfield(Exp,'nPoints'), Exp.nPoints = 1024; end
if ~isfield(Exp,'Harmonic'), Exp.Harmonic = 1; end
if ~isfield(Exp,'Range'), Exp.Range = []; end
if ~isfield(Exp,'ModAmp'), Exp.ModAmp = 0; end


if ~((Exp.Harmonic==0) || (Exp.Harmonic==1) || (Exp.Harmonic==2))
  error('Exp.Harmonic must be 0, 1 or 2.');
end

AutoRange = isempty(Exp.Range);
if ~AutoRange
  if (numel(Exp.Range)~=2) || (Exp.Range(1)>=Exp.Range(2))
    error('Exp.Range is invalid. It must be of the form [Bmin Bmax].');
  end
end

% Options
%-------------------------------------------------------------
if ~isfield(Opt,'AutoRangeZoom'), Opt.AutoRangeZoom = 1.5; end
if ~isfield(Opt,'reduce'), Opt.reduce = 1; end
if ~isfield(Opt,'Output'), Opt.Output = 'summed'; end
if ~isfield(Opt,'Verbosity'), Opt.Verbosity=0;end

separateOutput = strcmp(Opt.Output,'separate');

if ~isfield(Opt,'Method')
  Opt.Method = 2;
end
if ~isfield(Opt,'CalcMethod')
  Opt.CalcMethod = 'perturb1';
end


%
% Multicomponent
if iscell(Sys)
    if isfield(Sys{1},'mwFreq'), Exp.mwFreq = Sys{1}.mwFreq; end %trick for mwFreq vary
    spc=zeros(1,Exp.nPoints);
    for i=1:size(Sys,2)
        Sys1=Sys{i};
        [B,spc1]=exchange(Sys1,Exp,Opt);
        spc=spc+spc1;
    end
else
% 
%Isotopologues calculation
if isfield(Sys,'Nucs'),
    if isfield(Sys,'n'),Is=isotopologues(Sys.Nucs,Sys.n);
    else Is=isotopologues(Sys.Nucs); end
    nIso=max(size(Is));
else
    nIso=1;
end
if nIso>1
    if isfield(Sys,'Am')
    Sys.A=Sys.A(Sys.Am);
    Sys = rmfield(Sys,'Am');
    end
    Sys1=Sys;
    Sys1.A=Sys1.A.';
    Is=isotopologues(Sys1);
    spc=zeros(1,Exp.nPoints);
    for i=1:nIso
        Sys1=Is(:,i);
        
        if isfield(Sys1,'lbm')
            if iscell(Sys.lbm)
                Sys1.lbm=Sys.lbm{i};
            else Sys1.lbm=Sys.lbm;
            end
        end
        Sys1.A=Sys1.A';
        [B,spc1]=exchange(Sys1,Exp,Opt);
        spc=spc+spc1;
    end
else
    
    if isfield(Sys,'mwFreq'), Exp.mwFreq = Sys.mwFreq; end %trick for mwFreq vary
% Broadening
%------------------------------------------------------------
if isfield(Sys,'lbm')
    if ~isfield(Sys,'lb'), error ('Sys.lbm is given but Sys.lb is absent'); end
    if numel(Sys.lb)~=1, error ('Sys.lb must have single value if Sys.lbm is given'); end
    if ~iscell(Sys.lbm), Sys.lb = Sys.lb*Sys.lbm; end
end
  
% Spin system
%------------------------------------------------------------
if ~isfield(Sys,'g'), Sys.g = gfree; end
if isfield(Sys,'gm'), Sys.g = Sys.g(Sys.gm); end
if ~isfield(Sys,'Nucs'), Sys.Nucs = ''; end
if isfield(Sys,'weight'), weight=Sys.weight; end
if isfield(Exp,'Temperature'), weight=weight/Exp.Temperature; end %temperature scalling factor
if ~isempty(Sys.Nucs)
  I = nucspin(Sys.Nucs);
  if ~isfield(Sys,'n'), Sys.n = ones(size(I)); end
  
  if isfield(Sys,'Am')
    Sys.A=Sys.A(Sys.Am);
  end
  nNucGroups = size(Sys.A,2);
  if isfield(Sys,'xc')
    if size(Sys.xc,2)~=nNucGroups
      error('Sys.xc and Sys.A must have the same number of elements.');
    end
    for iSite = 1:size(Sys.xc,1)
      p = Sys.xc(iSite,:); % permutation
      if any(sort(p)~=1:nNucGroups)
        error('Row %d of Sys.xc is incorrect: this is not a permutation of the nuclei.',iSite);
      end
      if any(Sys.n~=Sys.n(p))
        error('Row %d of Sys.xc is incorrect: nuclei have different multiplicity and cannot be swapped.',iSite);
      end
      if any(I~=I(p))
        error('Row %d of Sys.xc is incorrect: nuclei are different and cannot be swapped',iSite);
      end
    end
    xc = [1:nNucGroups; Sys.xc]; % supplement identity permutation if absent
    xc = unique(xc,'rows'); % remove duplicate permutations
    Sys.A = Sys.A(xc);
    
  end
  if numel(I)~=nNucGroups
  error('Sys.A and Sys.Nucs are not consistent.');
  end
  nSites = size(Sys.A,1);
  if numel(Sys.g)~=nSites
    Sys.g = Sys.g(ones(1,nSites));
  end
else
  if isfield(Sys,'A')&&isempty(Sys.A),Sys=rmfield(Sys,'A'); end
  if isfield(Sys,'n')&&isempty(Sys.n),Sys=rmfield(Sys,'n'); end
  if isfield(Sys,'A'), error('Sys.A given, but Sys.Nucs is missing.'); end
  if isfield(Sys,'n'), error('Sys.n given, but Sys.Nucs is missing.'); end
  nNucGroups = 0;
  nSites = numel(Sys.g);
end

%if (nSites==1), error('Only one site given, no chemical exchange!'); end

% line width
if ~isfield(Sys,'lwpp'), Sys.lwpp = []; end
if ~isfield(Sys,'lw'), Sys.lw = []; end
if isempty(Sys.lw) && isempty(Sys.lwpp), error('Line width (Sys.lwpp or Sys.lw) is missing.'); end

if ~isempty(Sys.lw) && isempty(Sys.lwpp), 
    Sys.lwpp = Sys.lw/sqrt(3); 
end

if (Sys.lwpp<=0), error('Linewidth (Sys.lw, Sys.lwpp) must be positive.'); end
if (numel(Sys.lwpp)~=1) && (numel(Sys.lwpp)~=nSites)
  error('Either 1 or %d line widths (Sys.lw, Sys.lwpp) expected for %d sites.',nSites,nSites);
end
if isfield(Sys,'Pop')&&(numel(Sys.Pop))~=nSites
    error('%d populations expected for %d sites.',nSites,nSites);
end
lwpp = mt2mhz(Sys.lwpp,mean(Sys.g)); % field to frequency conversion
Gamma = lwpp*pi*sqrt(3); % peak-to-peak lw -> Gamma (= 1/T2) conversion

% Exchange matrix and rates
%----------------------------------------------------------------
if ~isfield(Sys,'k'), Sys.k = []; end
%if isempty(Sys.k), error('Sys.k (exchange rates/matrix) is missing.'); end
if isfield(Sys,'dH')&&~isfield(Sys,'dS'), error('dH is given but dS is absent'); end
if isfield(Sys,'dS')&&~isfield(Sys,'dH'), error('dS is given but dH is absent'); end
if isfield(Sys,'dH')&&isfield(Sys,'dS')&&~isfield(Exp,'Temperature'), error('dH and dS are given but Temperature is absent'); end
if isfield(Sys,'dH')&&isfield(Sys,'dS')&&isfield(Exp,'Temperature'),Sys.k=1e-6*boltzm*Exp.Temperature/planck*exp((-Sys.dH+Exp.Temperature*Sys.dS)/Exp.Temperature/molgas); end
if isfield(Sys,'km')
   if min(size(Sys.k))~=1,error('Sys.k must be linear list of rate constants.'); end
   if max(Sys.km)>max(size(Sys.k)), error('Sys.km contains invalid value'); end
   Sys.k=[0 Sys.k]; %force zero value
   Sys.km=Sys.km+1; %force correct indexing
   Sys.k = Sys.k(Sys.km); %building 2D matrix
end
k = Sys.k;

twoSiteMutualExchange = (nSites==2) && (numel(k)==1);

if min(size(k))==1||isempty(Sys.k)
  % mutual exchange: k contains linear list of rate constants
  nRates = nSites*(nSites-1)/2;
  if (numel(k)~=nRates)
    error('For %d-site mutual exchange, %d rate constants expected in Sys.k.',nSites,nRates);
  end
  if any(Sys.k<0), error('Sys.k cannot be negative.'); end
  X = zeros(nSites);
  X(tril(ones(nSites),-1)>0) = k;
  X = X + X.';
  X(1:nSites+1:end) = -sum(X);
else
  % general 2D exchange: k = general matrix for an arbitrary number of sites
  if (size(k,1)~=nSites) || (size(k,2)~=nSites) || numel(k)~=nSites^2
    error('2D exchange matrix in Sys.k has wrong size.');
  end
  X = k;
  if ~isfield(Sys,'Pop') %calculation of sites populations as formation-consumption ratio
      k(1:nSites+1:end) = 0; %force diagonal to zeroes
      P2 = sum(k,2); %formation rate as a sum of the row
      P1 = sum(k,1); %consumption rate as a sum of the column
      P1 = P1(:); %some tricks to form right dimension of the Sys.Pop
      Sys.Pop = sum(P2/P1,2); %calculation of sites populations
  end
  X(1:nSites+1:end) = X(1:nSites+1:end) - sum(X,1); % set diagonal so column sums are 0
end

% Spectral shifts, positions and intensities
%-------------------------------------------------------------

% Shifts in the limit of no exchange
for iGroup = 1:nNucGroups
  % first-order perturbation limit
  if I(iGroup)~=0
      Int{iGroup} = equivsplit(I(iGroup),Sys.n(iGroup));
  else
      Int{iGroup} = 1;
  end
  mFmax = Sys.n(iGroup)*I(iGroup);
  mF = -mFmax:+mFmax;
  for iSite = 1:nSites
    shift{iGroup,iSite} = mF*Sys.A(iSite,iGroup);
  end
end

% choose a static field
% (sim is done as a freq sweep at this static field)
if AutoRange
  B0 = mhz2mt(Exp.mwFreq*1e3,mean(Sys.g));
else
  B0 = mean(Exp.Range); % center of given sweep range
end
nu0 = mt2mhz(B0,Sys.g)-Exp.mwFreq*1e3; % MHz

% Positions and intensities in the static limit of no exchange
%--------------------------------------------------------------
if strcmp(Opt.CalcMethod,'perturb1')
    if (nNucGroups>0)
        for iSite = 1:nSites
            cumshift =  sum(allcombinations(shift{:,iSite}),2);
            nu(:,iSite) = nu0(iSite) + cumshift;
        end
        intensity = prod(allcombinations(Int{:}),2);
    else
         nu = nu0;
         intensity = 1;
    end
elseif strcmp(Opt.CalcMethod,'perturb2')
    if max(Sys.n)>1, error('Equivalent nuclei not supported, use perturb1 method or rebuild spin system'); end
    Par.Field=B0;
    Par.Range=Exp.Range;
    Op.PerturbOrder=2;
    for iSite = 1:nSites
        Sys1=rmfield(Sys,'lw');
        Sys1.g=Sys.g(iSite);
        Sys1.A=Sys.A(iSite,:);
        [nu1, intensity1]=resfreqs_perturb(Sys1,Par,Op);
        nu(:,iSite)=nu1-Exp.mwFreq*1e3;
        intensity(:,iSite)=intensity1;
    end
elseif strcmp(Opt.CalcMethod,'breitrabi')||strcmp(Opt.CalcMethod,'matrix')
    if max(Sys.n)>1, error('Equivalent nuclei not supported, use perturb1 method or rebuild spin system'); end
    gNucs=nucgval(Sys.Nucs);
    INucs=nucspin(Sys.Nucs);
    nNucs=numel(gNucs);
    for iSite = 1:nSites
    mwFreqs = Exp.mwFreq; %starting value
    giso = Sys.g(iSite);
    for i=1:nNucs
        gn = gNucs(i);
        I = INucs(i);
        mI = -I:I;
        aiso = Sys.A(iSite,i); % MHz
        aiso = aiso*1e6*planck; % MHz -> J
        B1=[];
        for j=1:max(size(mwFreqs))
            mwFreq=mwFreqs(j);
            Opt.MaxIterations = 15;
            Opt.Accuracy = 1e-12;
            B = 0;
            mwFreq = mwFreq*1e9; % MHz -> Hz
            gammae = giso*bmagn;
            gamman = 0; % to obtain start field, neglect NZI
            relChangeB = inf;
            for iter = 1:Opt.MaxIterations
                nnu = mwFreq + gamman/planck*B;
                q = 1 - (aiso./(2*planck*nnu)).^2;
                Bnew = aiso./(gammae+gamman)./q.*(-mI+sign(aiso)*sqrt(mI.^2+q.*((planck*nnu/aiso).^2 - (I+1/2)^2)));
                if ~isreal(Bnew)
                    error('Hyperfine coupling too large compared to microwave frequency.');
                end
                if (B~=0), relChangeB = 1 - Bnew./B; end
                B = Bnew;
                gamman = gn*nmagn; % re-include NZI
                if abs(relChangeB)<Opt.Accuracy, break; end
            end
            B=B*1000; %T->mT
            B1=[B1 B];
        end
        mwFreqs=mt2mhz(B1,giso)/1000; %GHz
    end
    if strcmp(Opt.CalcMethod,'breitrabi')
        nu1=mt2mhz(B0-B1,giso); %MHz
        nu1=nu1';
        nu(:,iSite)=nu1;
        intensity(:,iSite) =ones(numel(nu1),1); %no equiv nuclei
    else
        Sys1=rmfield(Sys,'lw');
        Sys1.g=Sys.g(iSite);
        Sys1.A=Sys.A(iSite,:);
        Par.mwFreq=Exp.mwFreq;
        Par.Range=Exp.Range;
        Op.Treshold=0.01;
        [B2,intensity1] = resfields(Sys1,Par,Op);
        if numel(B2)<numel(B1), 
           B1=B1(1:numel(B2));
        end
        if numel(B1)<numel(B2), 
           B2=B2(1:numel(B1));
        end
        Ii=[];
        for i=1:numel(B1);
            B=abs(B2-B1(i));
            while true              %equal fields resolving
                [C,I]=min(B);       %nearby field
                if sum(Ii==I)<1;    %current field is unique
                    Ii(i)=I;        %use it value
                    break
                end
                B(I)=max(B)+1;        %current field is not unique, use next one
        end
        end
        B2=B2(Ii);
        nu1=mt2mhz(B0-B2,giso); %MHz
        %nu1=nu1';
        nu(:,iSite)=nu1;
        intensity(:,iSite)=intensity1(Ii);
    end
    end
else
    error('Unknown calculation method');
end
intensity = intensity/sum(intensity);
maxshift = max(abs(nu(:)));
nLines = numel(intensity);
%linebroadening matrix
lb=zeros(nLines,nSites);
if isfield(Sys,'lb')
    if numel(Sys.lb)~=numel(lb), error ('lb matrix size mismatch, %d X %d is expected',nLines,nSites);end
    lb=pi*mt2mhz(Sys.lb,mean(Sys.g));
end

% Reduction
%-----------------------------------------------------------------
% For two-site mutual exchange, almost half of the line pairs
% occur twice (e.g. nu(k1,:) = [-10 7] and nu(k2,:) = [7 -10])
% Line pairs occure once only if for each group of nuclei 
% mF of site 1 is equal to mF of site 2 (that is, the line
% position does not change upon exchange).
%if twoSiteMutualExchange
if numel(Gamma)~=1||isfield(Sys,'lb'), Opt.reduce = 0; end %force disable reducing for different lws works properly  
  switch Opt.reduce
    case 0 % keep duplicates
      
    case 1 % remove duplicates
      %[nu intensity]
      nu = sort(nu,2);
      [nu,idx] = sortrows(nu);
      intensity = intensity(idx);
      
      keep = [true; any(abs(nu(1:end-1,:)-nu(2:end,:))>1e-5,2)];
      nu = nu(keep,:);
      if Opt.Verbosity==1,fprintf('reduction: %d -> %d (%0.3g%%)\n',nLines,size(nu,1),100*size(nu,1)/nLines);end
      
      iLine0 = 1;
      for iLine = 1:nLines
        if keep(iLine)
          iLine0 = iLine;
        else
          intensity(iLine0) = intensity(iLine0)+intensity(iLine);
        end
      end
      intensity = intensity(keep);
      nLines = size(nu,1);
      %[nu intensity]
      
    otherwise
      error('Unknown value for Opt.reduce')
  end


% Tresholding: remove lines with very low intensity
keep = intensity>0.001*max(intensity);
nu = nu(keep,:);
intensity = intensity(keep);
if Opt.Verbosity==1,fprintf('thresholding: %d -> %d (%0.3g%%)\n',nLines,size(nu,1),100*size(nu,1)/nLines);end
nLines = numel(intensity);

if AutoRange
  freqSweepRange = max(max(Gamma),maxshift)*Opt.AutoRangeZoom*2;
  Exp.Range = B0 + [-1 1]*mhz2mt(freqSweepRange/2,mean(Sys.g));
end
freqSweepRange = mt2mhz(diff(Exp.Range),mean(Sys.g));
nux = linspace(1,-1,Exp.nPoints)*freqSweepRange/2;

% Compute spectrum
%--------------------------------------------------------------
if separateOutput, nSpectra = nLines; else nSpectra = 1; end
spc = zeros(nSpectra,Exp.nPoints);

% Convert all frequenies to angular frequencies
nux = 2*pi*nux;
nu = 2*pi*nu;

switch (Opt.Method)
  case -3
    % Explicit method, two-site mutual exchange, Matlab accumulation
    %---------------------------------------------------------------
    dnu = nu(:,1) - nu(:,2);
    delta2 = (dnu/2).^2;
    numean = mean(nu,2);
    
    % Mex accumulation
    spc = exlinesum2(k,Gamma,numean,delta2,intensity,nux);
    
  case -2
    % Explicit method, two-site mutual exchange, mex accumulation
    %-------------------------------------------------------------
    dnu = nu(:,1) - nu(:,2);
    delta2 = (dnu/2).^2;
    numean = mean(nu,2);
    
    % Matlab accumulation
    a = 2*k+Gamma;
    c0 = Gamma*(2*k+Gamma);
    d0 = 2*(k+Gamma);
    for iLine = 1:nLines
      xi = nux - numean(iLine);
      b = xi;
      c = c0 + delta2(iLine) - xi.^2;
      d = d0*xi;
      spc_ = intensity(iLine)*(a*c+b.*d)./(c.^2+d.^2);
      if separateOutput
        spc(iLine,:) = spc_;
      else
        spc = spc + spc_;
      end
    end

  case -1
    % Liouville method, two-site mutual exchange, Matlab accumulation
    %----------------------------------------------------------------
    
    if ~twoSiteMutualExchange
      error('This method works only for two-site mutual exchange.');
    end
    if separateOutput
      error('Opt.Output=''separate'' not supported with this method.');
    end
      
    % S and L computation
    dnu = nu(:,1)-nu(:,2);
    q = sqrt(4*k^2-dnu.^2)/2;
    numean = mean(nu,2);
    Svalues = repmat(intensity,1,nSites).*[1-k./q, 1+k./q]/2;
    Lvalues = [-Gamma-k-q-1i*numean, -Gamma-k+q-1i*numean];
        
    % accumulation
    % treat special case 4k^2 == dnu using one-line expression
    idx = find(q==0);
    for ii = 1:numel(idx)
      zeta = Gamma+k+1i*(nux-numean(idx(ii)));
      spc = spc + intensity(idx(ii))*real((k+zeta)./zeta.^2);
    end

    % accumulation
    for iLine = 1:numel(Lvalues)
      S = Svalues(iLine);
      if isinf(S) || isnan(S), continue; end
      L = Lvalues(iLine);
      x = imag(L)+nux;
      spc_ = -(real(L)*real(S)+imag(S)*x)./(real(L)^2+x.^2);
      spc = spc + spc_;
    end
    
  case 0
    
    % Liouville method, two-site mutual exchange, mex accumulation
    %-------------------------------------------------------------
    
    if ~twoSiteMutualExchange
      error('This method works only for two-site mutual exchange.');
    end
    if separateOutput
      error('Opt.Output=''separate'' not supported with this method.');
    end
    
    % S and L computation
    dnu = abs(nu(:,1)-nu(:,2));
    q = sqrt(4*k^2-dnu.^2)/2;
    numean = mean(nu,2);
    Svalues = repmat(intensity,1,nSites).*[1-k./q, 1+k./q]/2;
    Lvalues = [-Gamma-k-q-1i*numean, -Gamma-k+q-1i*numean];
    
    % accumulation
    % (a) treat special case 4k^2==dnu using one-term expression
    %     (Svalues(:,iLine)==inf in this case)
    spc = zeros(1,Exp.nPoints);
    idx = find(q==0);
    for iLine = 1:numel(idx)
      zeta = Gamma+k+1i*(nux-numean(idx(iLine)));
      spc = spc + intensity(idx(iLine))*real((k+zeta)./zeta.^2);
    end
    
    % (b) general case: mex accumulation
    %     (omits lines with Svalues(:,iLine)==inf)
    spc = spc + exlinesum(Svalues,Lvalues,nux);
        
  case 1
    
    % Liouville method, general n-site exchange, mex accumulation
    %---------------------------------------------------------------
    if separateOutput
      error('Opt.Output=''separate'' not supported with this method.');
    end
    
    % S and L computation
    if isfield(Sys,'Pop')
        Pop = Sys.Pop(:)/sum(Sys.Pop); % fractional populations
        
    else
        Pop = ones(nSites,1)/nSites; % fractional populations
    end
    Det = ones(1,nSites); % detection operator
    if numel(Gamma)==1
        M0 = -Gamma*eye(nSites) + X;
    else
        M0 = diag(-Gamma) + X;
    end
    Lvalues = zeros(nSites,nLines);
    Svalues = zeros(nSites,nLines);
    for iLine = 1:nLines
      [U,Lambda] = eig(M0 - diag(lb(iLine,:)) - diag(1i*nu(iLine,:)));
      Lvalues(:,iLine) = diag(Lambda);
      Svalues(:,iLine) = (Det*U).'.*(U\Pop);
    end
    Svalues = Svalues.*repmat(intensity.',nSites,1);
    
    % accumulation
    spc = exlinesum(Svalues,Lvalues,nux);
    
  case 2
    
    % Liouville method, general n-site exchange, Matlab accumulation
    %---------------------------------------------------------------
    
    % S and L computation
    if isfield(Sys,'Pop')
        Pop = Sys.Pop(:)/sum(Sys.Pop); % fractional populations
    else
        Pop = ones(nSites,1)/nSites; % fractional populations
    end
    
    Det = ones(1,nSites);% % detection operator
    if numel(Gamma)==1
        M0 = -Gamma*eye(nSites) + X;
    else
        M0 = diag(-Gamma) + X;
    end
    Lvalues = zeros(nSites,nLines);
    Svalues = zeros(nSites,nLines);
    for iLine = 1:nLines
      
      [U,Lambda] = eig(M0 - diag(lb(iLine,:)) - diag(1i*nu(iLine,:)));
      Lvalues(:,iLine) = diag(Lambda);
      Svalues(:,iLine) = (Det*U).'.*(U\Pop);

    end
    Svalues = Svalues.*repmat(intensity.',nSites,1);
    
    % accumulation
    idx = 1;
    for iLine=1:nLines
      for iSite=1:nSites
        S = Svalues(iSite,iLine);
        L = Lvalues(iSite,iLine);
        x = imag(L)+nux;
        spc_ = -(real(L)*real(S)+imag(S)*x)./(real(L)^2+x.^2);
        if separateOutput 
              spc(idx,:) = spc_;
              idx = idx + 1;
          else
          spc = spc + spc_;
        end
      end
    end

 %   for iLine=1:nLines
 %     spc_ = 0;
 %     for iSite=1:nSites
 %       S = Svalues(iSite,iLine);
 %       L = Lvalues(iSite,iLine);
 %       x = imag(L)+nux;
 %       spc_ = spc_ -(real(L)*real(S)+imag(S)*x)./(real(L)^2+x.^2);
 %     end
 %     if separateOutput
 %       spc(iLine,:) = spc_;
 %       
 %     else
          
 %         spc = spc + spc_;
 %       end
  %  end
    
  case 3
    
    % Bloch/McConnell, two-site mutual exchange
    %-------------------------------------------------------------
    % my own version
    % (for another form see also Weil/Bolton p.320, Eq.(10.28b))
    
    if ~twoSiteMutualExchange
      error('Bloch/McConnel expressions implemented for two-site mutual exchange only.')
    end
    dnu = nu(:,1)-nu(:,2);
    q2 = k^2 - dnu.^2/4;
    numean = mean(nu,2);
    zeta0 =  Gamma + k + 1i*nux;
    for iLine = 1:nLines
      zeta = zeta0 - 1i*numean(iLine);
      Z_ = (k+zeta)./(q2(iLine)-zeta.^2);
      Z_ = -intensity(iLine)*real(Z_);
      if separateOutput
        spc(iLine,:) = Z_;
      else
        spc = spc + Z_;
      end
    end

  otherwise
    error('Unknown method.');
end

B = linspace(Exp.Range(1),Exp.Range(2),Exp.nPoints);


% Field modulation
%-------------------------------------------------------------
if (Exp.Harmonic>0)
  if (Exp.ModAmp==0)
    for m=1:Exp.Harmonic, spc = deriv(B,spc.').'; end
  else
    spc = fieldmod(B,spc,Exp.ModAmp,Exp.Harmonic);
  end
end

% Gaussian broadening
%-------------------------------------------------------------
if isfield(Sys,'lwg')
    df=(max(Exp.Range)-min(Exp.Range))/Exp.nPoints; %field step for convspec
    
    if min(size(spc))==1
        spc=convspec(spc,df,Sys.lwg,0,1);
    else
        spc=convspec(spc,df,[0 Sys.lwg],0,1);
    end
end
end
end
% Output
%-------------------------------------------------------------
spc=spc*weight;
switch nargout
  case 0
    plot(B,spc);
    title(sprintf('%0.8g GHz',Exp.mwFreq));
    axis tight
    xlabel('magnetic field (mT)');
  case 1
    varargout = {spc};
  case 2
    varargout = {B,spc};
  otherwise
    error('Wrong number of output arguments.');
end

function Combs = allcombinations(varargin)

if (nargin==0), Combs = []; return; end

Combs = varargin{1}(:);
nCombs = numel(Combs);

for iArg = 2:nargin
  New = varargin{iArg}(:);  nNew = numel(New);
  [idxNew,idxCombs] = find(ones(nNew,nCombs));
  Combs = [Combs(idxCombs(:),:), New(idxNew(:))];
  nCombs = nCombs*nNew;
end

return
